@@ -2,21 +2,17 @@ from abc import ABC
 import numpy as np
 from typing import Optional, Union, Tuple
 from time import time
+from pathlib import Path
 from .condition import Condition
 from .activation import Activation
 from .utils import rfunctions as functions
+from .thresholds import Thresholds
 import logging
 
 logger = logging.getLogger(__name__)
 
-try:
-    import pandas as pd
-except ImportError:
-    class pd:
-        DataFrame = None
-        Series = None
-
 
+# noinspection PyUnresolvedReferences
 class Rule(ABC):
 
     """An abstract Rule object.
@@ -44,6 +40,13 @@ class Rule(ABC):
     """
 
     LOCAL_ACTIVATION = True
+    THRESHOLDS = None
+    """Thresholds that the Rule must meet to be good. See `ruleskit.thresholds.Thresholds` for more details."""
+
+    @classmethod
+    def SET_THRESHOLDS(cls, path: Union[str, Path, "TransparentPath"], show=False):
+        """Set thresholds globally for all futur Rules"""
+        cls.THRESHOLDS = Thresholds(path, show)
 
     def __init__(
         self, condition: Optional[Condition] = None, activation: Optional[Activation] = None,
@@ -58,6 +61,9 @@ class Rule(ABC):
 
         self._condition = condition
         self._activation = activation
+        self._thresholds = Rule.THRESHOLDS
+        if self._activation is not None:
+            self.check_thresholds("coverage")
 
         self._coverage = None
         self._prediction = None
@@ -66,6 +72,42 @@ class Rule(ABC):
         self._time_calc_activation = -1
         self._time_predict = -1
 
+        self._good = True
+        self._bad_because = None
+
+    def set_thresholds(self, path: Union[str, Path, "TransparentPath"], show=False):
+        """Set thresholds for this rule only"""
+        self._thresholds = Thresholds(path, show)
+
+    def check_thresholds(self, attribute: Optional[str] = None) -> None:
+        """If `ruleskit.rule.Rule.THRESHOLDS` is specified, will check that this rule is good regarding those
+        thresholds, and set the flags *good* and *bad_because* accordingly
+
+        Parameters
+        ----------
+        attribute: Optional[str]
+            If specified, will only check the threshold of this rule attribute. If not, will test every rule attributes
+            for which a threshold is defined.
+        """
+
+        if Rule.THRESHOLDS is None:
+            return
+
+        if attribute is not None:
+            if not Rule.THRESHOLDS(attribute, self):
+                self._bad_because = attribute
+                self._good = False
+            return
+
+        for attribute in dir(self):
+            if attribute.startswith("__"):
+                continue
+            if not Rule.THRESHOLDS(attribute, self):
+                self._bad_because = attribute
+                self._good = False
+                return
+        logger.debug(f"Rule {self} is good")
+
     @property
     def coverage(self) -> float:
         if self._activation is not None:
@@ -125,9 +167,21 @@ class Rule(ABC):
         return None
 
     @property
-    def prediction(self) -> float:
+    def prediction(self) -> Union[str, float]:
         return self._prediction
 
+    @property
+    def thresholds(self) -> Thresholds:
+        return self._thresholds
+
+    @property
+    def good(self) -> bool:
+        return self._good
+
+    @property
+    def bad_because(self) -> str:
+        return self._bad_because
+
     @property
     def time_fit(self) -> float:
         """Profiling attribute. Time in seconds taken to fit the rule"""
@@ -204,14 +258,14 @@ class Rule(ABC):
         """A Rule's length is the number of features it talks about"""
         return len(self._condition)
 
-    def evaluate(self, xs: Union[pd.DataFrame, np.ndarray]) -> Activation:
+    def evaluate(self, xs: Union["pd.DataFrame", np.ndarray]) -> Activation:
         """Computes and returns the activation vector from an array of features.
 
         Parameters
         ----------
-        xs: Union[pd.DataFrame, np.ndarray]
+        xs: Union[pd:DataFrame, np.ndarray]
             The features on which the check whether the rule is activated or not. Must be a 2-D np.ndarray
-            or pd.DataFrame.
+            or pd:DataFrame.
 
         Returns
         -------
@@ -223,53 +277,59 @@ class Rule(ABC):
         return a
 
     # noinspection PyUnusedLocal
-    def fit(self, xs: Union[pd.DataFrame, np.ndarray], y: np.ndarray, **kwargs):
+    def fit(self, xs: Union["pd.DataFrame", np.ndarray] = None, y: np.ndarray = None, **kwargs):
         """Computes activation, and other criteria dependant on the nature of the daughter class of the Rule,
         for a given xs and y.
 
         Parameters
         ----------
-        xs: Union[pd.DataFrame, np.ndarray]
+        xs: Union[pd:DataFrame, np.ndarray]
             The features on which the check whether the rule is activated or not. Must be a 2-D np.ndarray
-            or pd.DataFrame.
+            or pd:DataFrame.
         y: np.ndarray
             The targets on which to evaluate the rule prediction, and possibly other criteria. Must be a 1-D np.ndarray.
         kwargs: dict
             Other arguments used by daughter class
         """
         t0 = time()
-        self.calc_activation(xs)
-        self.calc_attributes(xs, y, **kwargs)
+        if xs is not None:
+            self.calc_activation(xs)
+
+        if self._activation is None:
+            raise ValueError("If fitting without specifying xs, activation must have been computed already.")
+        self.calc_attributes(y, **kwargs)
         if self.prediction is None:
             raise ValueError("'fit' did not set 'prediction' : did you overload 'calc_attributes' correctly ?")
+
         self._time_fit = time() - t0
 
-    def calc_attributes(self, xs: Union[pd.DataFrame, np.ndarray], y: np.ndarray, **kwargs):
+    def calc_attributes(self, y: Union[np.ndarray, "pd.Series"], **kwargs):
         """Implement in daughter class. Must set self._prediction."""
         raise NotImplementedError("To implement in daughter class")
 
-    def calc_activation(self, xs: Union[pd.DataFrame, np.ndarray]):
+    def calc_activation(self, xs: Union["pd.DataFrame", np.ndarray]):
         """Uses self.evaluate to set self._activation.
 
         Parameters
         ----------
-        xs: Union[pd.DataFrame, np.ndarray]
+        xs: Union[pd:DataFrame, np.ndarray]
             The features on which the check whether the rule is activated or not. Must be a 2-D np.ndarray
-            or pd.DataFrame.
+            or pd:DataFrame.
         """
         t0 = time()
         self._activation = self.evaluate(xs)
         self._time_calc_activation = time() - t0
+        self.check_thresholds("coverage")
 
-    def predict(self, xs: Optional[Union[pd.DataFrame, np.ndarray]] = None) -> Union[np.ndarray, pd.Series]:
+    def predict(self, xs: Optional[Union["pd.DataFrame", np.ndarray]] = None) -> Union[np.ndarray, "pd.Series"]:
         """Returns the prediction vector. If xs is not given, will use existing activation vector.
         Will raise ValueError is xs is None and activation is not yet known.
 
         Parameters
         ----------
-        xs: Optional[Union[pd.DataFrame, np.ndarray]]
+        xs: Optional[Union[pd:DataFrame, np.ndarray]]
             The features on which the check whether the rule is activated or not. Must be a 2-D np.ndarray
-            or pd.DataFrame. If not specified the rule's activation vector must have been computed already.
+            or pd:DataFrame. If not specified the rule's activation vector must have been computed already.
 
         Returns
         -------
@@ -291,7 +351,7 @@ class Rule(ABC):
             to_ret = to_ret.astype(str)
         to_ret[act == 1] = self.prediction
         self._time_predict = time() - t0
-        if xs is not None and isinstance(xs, pd.DataFrame):
+        if xs is not None and not isinstance(xs, np.ndarray):
             return xs.__class__(index=xs.index, data=to_ret).squeeze()  # So not to requier pandas explicitly
         return to_ret
 
@@ -308,6 +368,7 @@ class Rule(ABC):
         return self._activation.get_correlation(other._activation) * sign
 
 
+# noinspection PyUnresolvedReferences
 class RegressionRule(Rule):
 
     """Rule applied on continuous target data."""
@@ -344,16 +405,14 @@ class RegressionRule(Rule):
     def time_calc_std(self):
         return self._time_calc_std
 
-    def calc_attributes(self, xs: Union[pd.DataFrame, np.ndarray], y: np.ndarray, **kwargs):
+    def calc_attributes(self, y: Union[np.ndarray, "pd.Series"], **kwargs):
         """Computes prediction, standard deviation, and regression criterion
         
         Parameters
         ----------
-        xs: Union[pd.DataFrame, np.ndarray]
-            The features on which the check whether the rule is activated or not. Must be a 2-D np.ndarray
-            or pd.DataFrame.
-        y: np.ndarray
-            The targets on which to evaluate the rule prediction, and possibly other criteria. Must be a 1-D np.ndarray.
+        y: Union[np.ndarray, pd.Series]
+            The targets on which to evaluate the rule prediction, and possibly other criteria. Must be a 1-D np.ndarray
+            or pd.Series.
         kwargs: dict
             Arguments for calc_regression_criterion
         """
@@ -362,50 +421,57 @@ class RegressionRule(Rule):
         prediction_vector = self.prediction * self.activation
         self.calc_criterion(prediction_vector, y, **kwargs)
 
-    def calc_prediction(self, y: np.ndarray):
+    def calc_prediction(self, y: [np.ndarray, "pd.Series"]):
         """Computes the mean of all activated points in target y and use it as prediction
         
         Parameters
         ----------
-        y: np.ndarray
-            The targets on which to evaluate the rule prediction, and possibly other criteria. Must be a 1-D np.ndarray.
+        y: [np.ndarray, pd.Series]
+            The targets on which to evaluate the rule prediction, and possibly other criteria. Must be a 1-D np.ndarray
+            or pd.Series
         """
         t0 = time()
         if self.activation is None:
             return None
         self._prediction = functions.conditional_mean(self.activation, y)
         self._time_calc_prediction = time() - t0
+        self.check_thresholds("prediction")
 
-    def calc_std(self, y: np.ndarray):
+    def calc_std(self, y: Union[np.ndarray, "pd.Series"]):
         """Computes the standard deviation of all activated points in target y
         
         Parameters
         ----------
-        y: np.ndarray
-            The targets on which to evaluate the rule prediction, and possibly other criteria. Must be a 1-D np.ndarray.
+        y: Union[np.ndarray, pd.Series]
+            The targets on which to evaluate the rule prediction, and possibly other criteria. Must be a 1-D np.ndarray
+            or pd.Series.
         """
         t0 = time()
         if self.activation is None:
             return None
         self._std = functions.conditional_std(self.activation, y)
         self._time_calc_std = time() - t0
+        self.check_thresholds("std")
 
-    def calc_criterion(self, p, y, **kwargs):
+    def calc_criterion(self, p: Union[np.ndarray, "pd.Series"], y: Union[np.ndarray, "pd.Series"], **kwargs):
         """
         Parameters
         ----------
-        p: np.ndarray
-            Prediction vector. Must be a 1-D np.ndarray.
-        y: np.ndarray
-            The targets on which to evaluate the rule prediction, and possibly other criteria. Must be a 1-D np.ndarray.
+        p: Union[np.ndarray, pd.Series]
+            Prediction vector. Must be a 1-D np.ndarray or pd.Series.
+        y: Union[np.ndarray, pd.Series]
+            The targets on which to evaluate the rule prediction, and possibly other criteria. Must be a 1-D np.ndarray
+            or pd.Series.
         kwargs: dict
             Arguments for calc_regression_criterion
         """
         t0 = time()
         self._criterion = functions.calc_regression_criterion(p, y, **kwargs)
         self._time_calc_criterion = time() - t0
+        self.check_thresholds("criterion")
 
 
+# noinspection PyUnresolvedReferences
 class ClassificationRule(Rule):
 
     """Rule applied on discret target data."""
@@ -435,43 +501,48 @@ class ClassificationRule(Rule):
     def criterion(self) -> float:
         return self._criterion
 
-    def calc_attributes(self, xs: Union[pd.DataFrame, np.ndarray], y: np.ndarray, **kwargs):
+    def calc_attributes(self, y: Union[np.ndarray, "pd.Series"], **kwargs):
         """
         Parameters
         ----------
-        xs: Union[pd.DataFrame, np.ndarray]
+        xs: Union[pd:DataFrame, np.ndarray]
             The features on which the check whether the rule is activated or not. Must be a 2-D np.ndarray
-            or pd.DataFrame.
-        y: np.ndarray
-            The targets on which to evaluate the rule prediction, and possibly other criteria. Must be a 1-D np.ndarray.
+            or pd:DataFrame.
+        y: Union[np.ndarray, pd.Series]
+            The targets on which to evaluate the rule prediction, and possibly other criteria. Must be a 1-D np.ndarray
+            or pd.Series.
         kwargs: dict
             Arguments for calc_classification_criterion
         """
         self.calc_prediction(y)
         self.calc_criterion(y, **kwargs)
 
-    def calc_prediction(self, y: np.ndarray):
+    def calc_prediction(self, y: [np.ndarray, "pd.Series"]):
         """
         Parameters
         ----------
-        y: np.ndarray
-            The targets on which to evaluate the rule prediction, and possibly other criteria. Must be a 1-D np.ndarray.
+        y: [np.ndarray, pd.Series]
+            The targets on which to evaluate the rule prediction, and possibly other criteria. Must be a 1-D np.ndarray
+            or pd.Series.
         """
         t0 = time()
         if self.activation is None:
             raise ValueError("The activation vector has not been computed yet.")
         self._prediction = functions.most_common_class(self.activation, y)
         self._time_calc_prediction = time() - t0
+        self.check_thresholds("prediction")
 
-    def calc_criterion(self, y, **kwargs):
+    def calc_criterion(self, y: Union[np.ndarray, "pd.Series"], **kwargs):
         """
         Parameters
         ----------
-        y: np.ndarray
-            The targets on which to evaluate the rule prediction, and possibly other criteria. Must be a 1-D np.ndarray.
+        y: Union[np.ndarray, pd.Series]
+            The targets on which to evaluate the rule prediction, and possibly other criteria. Must be a 1-D np.ndarray
+            or pd.Series
         kwargs: dict
             Arguments for calc_classification_criterion
         """
         t0 = time()
         self._criterion = functions.calc_classification_criterion(self.activation, self.prediction, y, **kwargs)
         self._time_calc_criterion = time() - t0
+        self.check_thresholds("criterion")
